/******************************************************************************
*
* (c) Copyright 2015, Freescale Semiconductor Inc.
*
***************************************************************************//*!
/*****************************************************************************/
/*                                                                           */
/* FILE NAME     :  main_core1.c                                             */
/* DESCRIPTION   :  eTPU Engine Control Demo                                 */
/* DATE          :  July 1, 2015                                            */
/*                                                                           */
/*****************************************************************************/

#include "../headers/project.h"
#include "../headers/mpc5746r.h"
#include "../etpu/etpu_gct.h"            /* eTPU Configuration */
#include "../etpu/utils/etpu_util.h"     /* General C Functions for the eTPU */
#include "../etpu/crank/etpu_crank.h"    /* eTPU CRANK API */
#include "../etpu/cam/etpu_cam.h"        /* eTPU CAM API */

#include "../FreeMASTER/freemaster.h"

extern void Clock_and_Mode_Init(void);

/******************************************************************************
* Global variables
******************************************************************************/
/* eTPU Engine A load as a percentage */
uint32_t etpu_engine_load;
/* eTPU log arrays */
uint24_t etpu_cam_log[CAM_LOG_SIZE];
uint24_t etpu_tooth_period_log[TEETH_PER_CYCLE];

/******************************************************************************
* FreeMASTER TSA tables
******************************************************************************/
/*
 * With TSA enabled, the user describes the global and static variables using 
 * so-called TSA tables. There can be any number of tables defined in 
 * the project files. Each table does have the identifier which should be
 * unique across the project. 
 *
 * Note that you can declare variables as Read-Only or Read-Write.
 * The FreeMASTER driver denies any write access to the Read-Only variables
 * when TSA_SAFETY is enabled.
 */

FMSTR_TSA_TABLE_BEGIN(fmstr_tsa_table_etpu_load)
    FMSTR_TSA_RO_VAR(etpu_engine_load, FMSTR_TSA_UINT32)
FMSTR_TSA_TABLE_END()

FMSTR_TSA_TABLE_BEGIN(fmstr_tsa_table_etpu_logs)
    FMSTR_TSA_RO_VAR(etpu_cam_log, FMSTR_TSA_UINT32)
    FMSTR_TSA_RO_VAR(etpu_tooth_period_log, FMSTR_TSA_UINT32)
FMSTR_TSA_TABLE_END()

/*
 * This list describes all TSA tables which should be exported to the 
 * FreeMASTER application.
 */
FMSTR_TSA_TABLE_LIST_BEGIN()
    FMSTR_TSA_TABLE(fmstr_tsa_table_etpu_load)
    FMSTR_TSA_TABLE(fmstr_tsa_table_etpu_logs)
    FMSTR_TSA_TABLE(fmstr_tsa_table_etpu_scaling)
    FMSTR_TSA_TABLE(fmstr_tsa_table_crank)
    FMSTR_TSA_TABLE(fmstr_tsa_table_cam)
    FMSTR_TSA_TABLE(fmstr_tsa_table_spark)
    FMSTR_TSA_TABLE(fmstr_tsa_table_fuel)
    FMSTR_TSA_TABLE(fmstr_tsa_table_inj)
    FMSTR_TSA_TABLE(fmstr_tsa_table_knock)
    FMSTR_TSA_TABLE(fmstr_tsa_table_tg)
FMSTR_TSA_TABLE_LIST_END()

/******************************************************************************
* Local function prototypes
******************************************************************************/
void fmpll_init(void);
void esci_a_init(void);
void LINFlex_init(void);
uint32_t get_etpu_load_a(void);

/******************************************************************************
* Interrupt handlers
******************************************************************************/
void etpu_crank_isr(void);
void etpu_cam_isr(void);
void etpu_fuel_1_isr(void);
void etpu_fuel_2_isr(void);
void etpu_fuel_3_isr(void);
void etpu_fuel_4_isr(void);
void etpu_spark_1_isr(void);
void etpu_spark_2_isr(void);
void etpu_spark_3_isr(void);
void etpu_spark_4_isr(void);
void etpu_knock_1_isr(void);
void etpu_knock_2_isr(void);
void etpu_inj_1_isr(void);
void etpu_inj_2_isr(void);
void etpu_inj_3_isr(void);
void etpu_inj_4_isr(void);
void etpu_tg_isr(void);



/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel CRANK
*
* @return  N/A
*
* @note    This interrupt is generated every time the engine position state
*          is changed. When the engine position state is 
*          FS_ETPU_ENG_POS_PRE_FULL_SYNC the logged Cam patern is decoded
*          in order to set tcr2_adjustment and achieve FULL_SYNC.
* 
******************************************************************************/
void etpu_crank_isr(void)
{
  uint24_t tcr2_adjustment;

  fs_etpu_clear_chan_interrupt_flag(ETPU_CRANK_CHAN);
 
  /* Follow Engine Position state */
  fs_etpu_crank_get_states(&crank_instance, &crank_states);
  switch(crank_states.eng_pos_state)
  {
  case FS_ETPU_ENG_POS_SEEK:
    /* Crank has stalled. Read Crank error to know the reason. */
    break;
  case FS_ETPU_ENG_POS_FIRST_HALF_SYNC:
    /* Crank has found the gap, or
       Crank did not received CRANK_HSR_SET_SYNC and hence Crank reset 
       the Cam log to repeat recording the Cam log. */
    break;
  case FS_ETPU_ENG_POS_PRE_FULL_SYNC:
    /* Cam signal is logged, the log corresponds to a segment of 
       teeth_per_sync Crank teeth starting from a gap.
       NOW THE CPU MUST RECOGNIZE THE LOGGED CAM PATTERN AND 
       1) SET tcr2_adjustment
       2) ASSIGN CRANK HSR = CRANK_HSR_SET_SYNC. */
    fs_etpu_cam_get_states(&cam_instance, &cam_states);
    fs_etpu_cam_copy_log(&cam_instance, &etpu_cam_log[0]);
    
    if((cam_states.log_idx == 3) && (etpu_cam_log[0] & 0x01000000))
    {
      /* 3 transition logged,
         the first transition is a rising one 
         => the  first half-cycle was logged */
      /* Set, what angle should have been at the last gap */
      tcr2_adjustment = DEG2TCR2(360);
    }
    else if((cam_states.log_idx == 3) && ~(etpu_cam_log[0] & 0x01000000))
    {
      /* 3 transitions logged,
         the first transition is a falling one
         => the second half-cycle was logged */
      /* Set, what angle should have been at the last gap */
      tcr2_adjustment = DEG2TCR2(0);
    }
    else
    {
      /* Cam pattern is not recognized */
      break;
    }
    fs_etpu_crank_set_sync(&crank_instance, tcr2_adjustment);
    break;
  case FS_ETPU_ENG_POS_FULL_SYNC:
    /* Regular interrupt on the first tooth every engine cycle. */
    /* Clear errors */
    crank_states.error = 0;
    cam_states.error = 0;
    break;
  }

  /* Interface CRANK eTPU function */
  fs_etpu_crank_get_states(&crank_instance, &crank_states);
  fs_etpu_crank_config(&crank_instance, &crank_config);
  fs_etpu_crank_copy_tooth_period_log(&crank_instance, &etpu_tooth_period_log[0]);
  /* Interface CAM eTPU function */
  fs_etpu_cam_get_states(&cam_instance, &cam_states);
  fs_etpu_cam_config(&cam_instance, &cam_config);
  fs_etpu_cam_copy_log(&cam_instance, &etpu_cam_log[0]);

  /* Evaluate eTPU load */
  etpu_engine_load = get_etpu_load_a();
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel CAM
*
* @return  N/A
*
* @note    This interrupt is generated on detecting an error condition.
* 
******************************************************************************/
void etpu_cam_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_CAM_CHAN);
    /* Interface CAM eTPU function */
    fs_etpu_cam_get_states(&cam_instance, &cam_states);
    fs_etpu_cam_config(&cam_instance, &cam_config);
    fs_etpu_cam_copy_log(&cam_instance, &etpu_cam_log[0]);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel FUEL_1
*
* @return  N/A
*
* @note    This interrupt is generated at each stop angle. The Fuel 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_fuel_1_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_FUEL_1_CHAN);
    /* Interface FUEL eTPU functions */
    fs_etpu_fuel_get_states(&fuel_1_instance, &fuel_1_states);
    fs_etpu_fuel_config(&fuel_1_instance, &fuel_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel FUEL_2
*
* @return  N/A
*
* @note    This interrupt is generated at each stop angle. The Fuel 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_fuel_2_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_FUEL_2_CHAN);
    /* Interface FUEL eTPU functions */
    fs_etpu_fuel_get_states(&fuel_2_instance, &fuel_2_states);
    fs_etpu_fuel_config(&fuel_2_instance, &fuel_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel FUEL_3
*
* @return  N/A
*
* @note    This interrupt is generated at each stop angle. The Fuel 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_fuel_3_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_FUEL_3_CHAN);
    /* Interface FUEL eTPU functions */
    fs_etpu_fuel_get_states(&fuel_3_instance, &fuel_3_states);
    fs_etpu_fuel_config(&fuel_3_instance, &fuel_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel FUEL_4
*
* @return  N/A
*
* @note    This interrupt is generated at each stop angle. The Fuel 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_fuel_4_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_FUEL_4_CHAN);
    /* Interface FUEL eTPU functions */
    fs_etpu_fuel_get_states(&fuel_4_instance, &fuel_4_states);
    fs_etpu_fuel_config(&fuel_4_instance, &fuel_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel SPARK_1
*
* @return  N/A
*
* @note    This interrupt is generated at each recalc angle. The Spark 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_spark_1_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_SPARK_1_CHAN);
    /* Interface SPARK eTPU function */
    fs_etpu_spark_get_states(&spark_1_instance, &spark_1_states);
    fs_etpu_spark_config(&spark_1_instance, &spark_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel SPARK_2
*
* @return  N/A
*
* @note    This interrupt is generated at each recalc angle. The Spark 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_spark_2_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_SPARK_2_CHAN);
    /* Interface SPARK eTPU function */
    fs_etpu_spark_get_states(&spark_2_instance, &spark_2_states);
    fs_etpu_spark_config(&spark_2_instance, &spark_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel SPARK_3
*
* @return  N/A
*
* @note    This interrupt is generated at each recalc angle. The Spark 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_spark_3_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_SPARK_3_CHAN);
    /* Interface SPARK eTPU function */
    fs_etpu_spark_get_states(&spark_3_instance, &spark_3_states);
    fs_etpu_spark_config(&spark_3_instance, &spark_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel SPARK_4
*
* @return  N/A
*
* @note    This interrupt is generated at each recalc angle. The Spark 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_spark_4_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_SPARK_4_CHAN);
    /* Interface SPARK eTPU function */
    fs_etpu_spark_get_states(&spark_4_instance, &spark_4_states);
    fs_etpu_spark_config(&spark_4_instance, &spark_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel KNOCK_1
*
* @return  N/A
*
* @note    This interrupt is generated at each window end. The Knock 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_knock_1_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_KNOCK_1_CHAN);
    /* Interface KNOCK eTPU function */;
    fs_etpu_knock_config(&knock_1_instance, &knock_1_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel KNOCK_2
*
* @return  N/A
*
* @note    This interrupt is generated at each window end. The Knock 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_knock_2_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_KNOCK_2_CHAN);
    /* Interface KNOCK eTPU function */
    fs_etpu_knock_config(&knock_2_instance, &knock_2_config);
}

/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel INJ_1
*
* @return  N/A
*
* @note    This interrupt is generated before the start of injection sequence
*          on INJ 1 channel. The injection sequence parameters can be adjusted.
* 
******************************************************************************/
void etpu_inj_1_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_INJ_1_CHAN);
    /* Interface INJ eTPU function */
    fs_etpu_inj_get_states(&inj_1_instance, &inj_1_states);
    fs_etpu_inj_config(&inj_1_instance, &inj_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel INJ_2
*
* @return  N/A
*
* @note    This interrupt is generated before the start of injection sequence
*          on INJ 1 channel. The injection sequence parameters can be adjusted.
* 
******************************************************************************/
void etpu_inj_2_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_INJ_2_CHAN);
    /* Interface INJ eTPU function */
    fs_etpu_inj_get_states(&inj_2_instance, &inj_2_states);
    fs_etpu_inj_config(&inj_2_instance, &inj_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel INJ_3
*
* @return  N/A
*
* @note    This interrupt is generated before the start of injection sequence
*          on INJ 1 channel. The injection sequence parameters can be adjusted.
* 
******************************************************************************/
void etpu_inj_3_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_INJ_3_CHAN);
    /* Interface INJ eTPU function */
    fs_etpu_inj_get_states(&inj_3_instance, &inj_3_states);
    fs_etpu_inj_config(&inj_3_instance, &inj_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel INJ_4
*
* @return  N/A
*
* @note    This interrupt is generated before the start of injection sequence
*          on INJ 1 channel. The injection sequence parameters can be adjusted.
* 
******************************************************************************/
void etpu_inj_4_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_INJ_4_CHAN);
    /* Interface INJ eTPU function */
    fs_etpu_inj_get_states(&inj_4_instance, &inj_4_states);
    fs_etpu_inj_config(&inj_4_instance, &inj_config);
}


/***************************************************************************//*!
*
* @brief   Interrupt from eTPU channel TG
*
* @return  N/A
*
* @note    This interrupt is generated in each gap. The Tooth Generator 
*          parameters can be adjusted.
* 
******************************************************************************/
void etpu_tg_isr(void)
{
    /* Clear interrupt flag */
    fs_etpu_clear_chan_interrupt_flag(ETPU_TG_CRANK_CHAN);
    /* Interface TG eTPU function */
    fs_etpu_tg_get_states(&tg_instance, &tg_states);
    fs_etpu_tg_config(&tg_instance, &tg_config);
}

/***************************************************************************//*!
*
* @brief   SIUL2 configuration for eTPU input/output
*
* @return  N/A
* 
******************************************************************************/
void SIUL2_Configure(void)
{
    /* Setup eTPU A inputs */
    SIUL2.MSCR512_955[512-512].R = 9;  // eTPU A ch0 = PH[11]
    SIUL2.MSCR0_255[123].R = 0x008A0000;  // PH[11] -> PH[11] - Cam
    SIUL2.MSCR512_955[514-512].R = 9;  // eTPU A ch2 = PH[13]
    SIUL2.MSCR0_255[125].R = 0x008A0000;  // PH[13] -> PH[13] - Crank
    /* Setup eTPU A outputs */
    SIUL2.MSCR0_255[124].R = 0x12380001;  // ch1 = PH[12] -> PH[12] - TG Cam
    SIUL2.MSCR0_255[126].R = 0x12380001;  // ch3 = PH[14] -> PH[14] - TG Crank
    SIUL2.MSCR0_255[127].R = 0x12380001;  // ch4 = PH[15] -> PH[15] - Spark 1
    SIUL2.MSCR0_255[128].R = 0x12380001;  // ch5 = PI[0] -> PI[0] - Spark 2
    SIUL2.MSCR0_255[129].R = 0x12380001;  // ch6 = PI[1] -> PI[1] - Spark 3
    SIUL2.MSCR0_255[130].R = 0x12380001;  // ch7 = PI[2] -> PI[2] - Spark 4
    SIUL2.MSCR0_255[131].R = 0x12380001;  // ch8 = PI[3] -> PI[3] - Fuel 1
    SIUL2.MSCR0_255[132].R = 0x12380001;  // ch9 = PI[4] -> PI[4] - Fuel 2
    SIUL2.MSCR0_255[133].R = 0x12380001;  // ch10 = PI[5] -> PI[5] - Fuel 3
    SIUL2.MSCR0_255[101].R = 0x12380001;  // ch11 = PG[5] -> PE[6] - Fuel 4
    SIUL2.MSCR0_255[122].R = 0x12380001;  // ch12 = PH[10] -> PH[10] - Knock 1
    SIUL2.MSCR0_255[53].R = 0x12380001;  // ch13 = PD[5] -> PL[5] - Knock 2
    SIUL2.MSCR0_255[54].R = 0x12380001;  // ch14 = PD[6] -> PL[6] - Inj Bank 1
    SIUL2.MSCR0_255[55].R = 0x12380001;  // ch15 = PD[7] -> PL[7] - Inj Bank 2
    SIUL2.MSCR0_255[106].R = 0x12380001;  // ch16 = PG[10] -> PG[10] - Inj 1
    SIUL2.MSCR0_255[57].R = 0x12380001;  // ch17 = PD[9] -> PL[9] - Inj 2
    SIUL2.MSCR0_255[58].R = 0x12380001;  // ch18 = PD[10] -> PL[10] - Inj 3
    SIUL2.MSCR0_255[59].R = 0x12380001;  // ch19 = PD[11] -> PL[11] - Inj 4
}


/*****************************************************************************/
/*         Function: main                                                    */
/*         Content:  main function for core 1                                */
/*         Argument: void                                                    */
/*         Return value: void                                                */
/*****************************************************************************/

int main(void)
{
    volatile uint32_t i;

    /* Initialize LINFlex, UART */
    LINFlex_init();
    
    /* Confifure SIUL2 eTPU output and input*/
    SIUL2_Configure();

    /* Initialize eTPU */
    my_system_etpu_init();

    /* Initialize FreeMASTER */
    FMSTR_Init();
    
    /* Initialize interrupts */
    Irq_Init_Core0(); // - not necessary
    Irq_Init_Core1();

    /* Start eTPU */
    my_system_etpu_start();
    get_etpu_load_a();

    /* Main While Loop */
    while(1)
    {
	/* Set Fuel injection time - the value is updated in by FreeMASTER */
	fs_etpu_fuel_update_injection_time(&fuel_1_instance, &fuel_config);

	/* Interface TG eTPU function - this sets engine speed updated by FreeMASTER */
	fs_etpu_tg_get_states(&tg_instance, &tg_states);
	fs_etpu_tg_config(&tg_instance, &tg_config);

	/* update Crank and Cam latest states to see them in FreeMaster*/
	fs_etpu_crank_get_states(&crank_instance, &crank_states);
	fs_etpu_cam_get_states(&cam_instance, &cam_states);

	/* FreeMASTER processing on background */
	FMSTR_Poll();
    }
}
/***************************************************************************//*!
*
* @brief    Init LINFlex #2 to 115kbd @ 80MHz, for FreeMASTER.
*
* @return  N/A
*
******************************************************************************/
void LINFlex_init(void)
{
    //set protocol frequency from PLL0 with divider 2
    MC_CGM.AC0_DC4.R = 0x80010000;

    // SIUL2 configuration for LINFlex #2	// Port Pin	Number	MSCR_SSS	Function	Description	
    SIUL2.MSCR0_255[80].R = 0x32810009;		// PF[0]	80	0000_1001	LIN2TX		LINFlexD_2 Transmit Data Output
    SIUL2.MSCR0_255[81].R = 0x00080000;		// PF[1]	81	0000_0000	GPIO[81]	General Purpose	
    SIUL2.MSCR512_955[754 - 512].R = 0x00000002;// PF[1]	754	0000_0010	LIN2RX		LINFlexD_2 Receive Data Input

    // UART LINFlex #2 Configuration
    LINFlexD_2.LINCR1.B.SLEEP = 0x0;	// exit sleep mode
    LINFlexD_2.LINCR1.B.INIT = 0x1;		// Put LINFlex hardware in init mode  
    LINFlexD_2.UARTCR.B.UART = 0x1;		// Enable UART mode 
    LINFlexD_2.UARTCR.B.WL0 = 0x1;		// 8 bit Word Length in UART mode
    LINFlexD_2.UARTCR.B.WL1 = 0x0;		// 8 bit Word Length in UART mode
    LINFlexD_2.UARTCR.B.PCE = 0x0;		// no Parity
    LINFlexD_2.UARTCR.B.TxEn = 0x1;		// Enable transmission of data now
    LINFlexD_2.UARTCR.B.TDFLTFC = 0x0;	// Transmitter Data Field Length/TX FIFO Counter = 1 byte

    LINFlexD_2.LINIBRR.B.IBR = 0x00002B;	// LDIV Mantissa baud rate divider component (Integer Baud Rate Register - IBR)
    LINFlexD_2.LINFBRR.B.FBR = 0x6;		// Fraction baud rate divider component (Fractional Baud Rate Register - FBR)

    LINFlexD_2.LINCR1.B.INIT = 0x0;		// Put LINFlex hardware in normal mode

}
/***************************************************************************//*!
*
* @brief   Evaluate eTPU engine A load.
*
* @warning - This function is applicable on eTPU2 only.
*          - The first call of this function does not return a correct result.
*          - This function presumes the TCR1 clock is configured for system 
*            clock div 1.
*
* @return  eTPU engine A load percentage (0 - 100).
*
******************************************************************************/
uint32_t get_etpu_load_a(void)
{
  static uint24_t tcr1;
         uint24_t tcr1_last;
         uint24_t idle_cnt;
         uint24_t time_cnt;

  tcr1_last = tcr1;
  tcr1 = eTPU->TB1R_A.R;
  idle_cnt = fs_etpu_get_idle_cnt_a();
  fs_etpu_clear_idle_cnt_a();

  time_cnt = 0xFFFFFF & (tcr1 - tcr1_last);
  return(100*(time_cnt - idle_cnt)/time_cnt);
}


/*******************************************************************************
 *
 * Copyright:
 *  Freescale Semiconductor, INC. All Rights Reserved.
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Freescale
 *  Semiconductor, Inc. This software is provided on an "AS IS"
 *  basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, Freescale
 *  Semiconductor DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED,
 *  INCLUDING IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 *  PARTICULAR PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH
 *  REGARD TO THE SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF)
 *  AND ANY ACCOMPANYING WRITTEN MATERIALS.
 *
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL Freescale Semiconductor BE LIABLE FOR ANY DAMAGES WHATSOEVER
 *  (INCLUDING WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
 *  BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER
 *  PECUNIARY LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.
 *
 *  Freescale Semiconductor assumes no responsibility for the
 *  maintenance and support of this software
 ******************************************************************************/
